package net.knarcraft.ffmpegconverter.converter;

import net.knarcraft.ffmpegconverter.FFMpegConvert;
import net.knarcraft.ffmpegconverter.container.FFMpegCommand;
import net.knarcraft.ffmpegconverter.container.StreamProbeResult;
import net.knarcraft.ffmpegconverter.converter.module.DeInterlaceModule;
import net.knarcraft.ffmpegconverter.converter.module.ModuleExecutor;
import net.knarcraft.ffmpegconverter.handler.AvailableHardwareEncoderHandler;
import net.knarcraft.ffmpegconverter.streams.StreamObject;
import net.knarcraft.ffmpegconverter.utility.FFMpegHelper;
import net.knarcraft.ffmpegconverter.utility.FileHelper;
import net.knarcraft.ffmpegconverter.utility.FileUtil;
import net.knarcraft.ffmpegconverter.utility.OutputUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Implements all methods which can be useful for any implementation of a converter.
 */
public abstract class AbstractConverter implements Converter {

    final boolean debug = FFMpegConvert.getConfiguration().isDebugEnabled();
    private final String newExtension;
    protected String ffprobePath;
    protected String ffmpegPath;
    protected List<String> audioFormats;
    protected List<String> videoFormats;
    protected List<String> subtitleFormats;
    protected AvailableHardwareEncoderHandler encoderHandler = null;


    /**
     * Initializes variables used by the abstract converter
     */
    AbstractConverter(@Nullable String newExtension) {
        this.newExtension = newExtension;
        try {
            this.audioFormats = FileHelper.readLines(FileHelper.getBufferedReaderForInternalFile("/audio_formats.txt"));
            this.videoFormats = FileHelper.readLines(FileHelper.getBufferedReaderForInternalFile("/video_formats.txt"));
            this.subtitleFormats = FileHelper.readLines(FileHelper.getBufferedReaderForInternalFile("/subtitle_formats.txt"));
        } catch (IOException e) {
            OutputUtil.println("Unable to read audio and/or video formats from internal files.");
            System.exit(1);
        }
    }

    @Override
    public void convert(@NotNull File file) throws IOException {
        StreamProbeResult probeResult = FFMpegHelper.probeFile(this.ffprobePath, file, this.subtitleFormats);
        if (probeResult.parsedStreams().isEmpty()) {
            throw new IllegalArgumentException("The file has no valid streams. Please make sure the file exists and" +
                    " is not corrupt.");
        }

        String outExtension = this.newExtension != null ? this.newExtension : FileUtil.getExtension(file.getName());
        String newPath = FileUtil.getNonCollidingPath(file.getParentFile(), file, outExtension);
        OutputUtil.println();
        OutputUtil.println("Preparing to start process...");
        OutputUtil.println("Converting " + file);

        FFMpegCommand ffMpegCommand = generateConversionCommand(this.ffmpegPath, probeResult, newPath);
        // If the command is null, that means the file does not need conversion
        if (ffMpegCommand == null) {
            return;
        }

        // De-interlace the video if enabled
        if (FFMpegConvert.getConfiguration().deInterlaceVideo() &&
                (outExtension.equalsIgnoreCase("mkv") || outExtension.equalsIgnoreCase("mp4"))) {
            new ModuleExecutor(ffMpegCommand, List.of(new DeInterlaceModule())).execute();
        }

        String[] command = ffMpegCommand.getResult();
        // If no commands were given, no conversion is necessary
        if (command.length == 0) {
            return;
        }

        ProcessBuilder processBuilder = new ProcessBuilder(command);
        int exitCode = FFMpegHelper.runProcess(processBuilder, file.getParentFile(), "\n", true).exitCode();
        if (exitCode != 0) {
            handleError(ffMpegCommand, file, newPath);
        }
    }

    /**
     * Handles an ffmpeg conversion error
     *
     * @param ffMpegCommand <p>The failed ffmpeg command</p>
     * @param file          <p>The file that was to be converted</p>
     * @param newPath       <p>The path of the output file</p>
     * @throws IOException <p>If unable to produce output</p>
     */
    private void handleError(@NotNull FFMpegCommand ffMpegCommand, @NotNull File file,
                             @NotNull String newPath) throws IOException {
        File outputFile = new File(newPath);
        if (outputFile.exists() && !outputFile.delete()) {
            OutputUtil.println("Failed to remove failed output file. Please remove it manually");
        }

        String outputCodec = ffMpegCommand.getOutputVideoCodec();
        if (outputCodec != null && outputCodec.contains("_")) {
            String[] parts = outputCodec.split("_");
            String hardwareAcceleration = parts[parts.length - 1];
            List<String> encodingMethods = getAvailableHardwareEncodingMethods();
            if (encodingMethods.contains(hardwareAcceleration)) {
                OutputUtil.println("Disabling " + hardwareAcceleration + " hardware acceleration");
                encoderHandler.removeHardwareEncoding(hardwareAcceleration);
                encoderHandler.save();
                convert(file);
                return;
            }
        }
        throw new IllegalArgumentException("Conversion failed. Please check the preceding output.");
    }

    /**
     * Gets the available methods for hardware encoding
     *
     * @return <p>Available hardware encoding methods</p>
     */
    @NotNull
    protected List<String> getAvailableHardwareEncodingMethods() {
        if (encoderHandler == null) {
            encoderHandler = AvailableHardwareEncoderHandler.load();
            if (encoderHandler.availableHardwareEncodings().isEmpty()) {
                List<String> hardwareEncoding;
                try {
                    hardwareEncoding = new ArrayList<>(FFMpegHelper.getHWAcceleration(ffmpegPath));
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
                hardwareEncoding.remove(0);
                encoderHandler = new AvailableHardwareEncoderHandler(hardwareEncoding);
                encoderHandler.save();
            }
        }

        return encoderHandler.availableHardwareEncodings();
    }

    /**
     * Sets the output indexes for the given streams
     *
     * <p>This will basically mark the given streams' order as the order they will appear in the output file. This is
     * used when specifying specific codecs per-stream in the output file.</p>
     *
     * @param streams <p>The streams to set the output indexes for</p>
     */
    protected <K extends StreamObject> void setOutputIndexes(@NotNull List<K> streams) {
        for (int i = 0; i < streams.size(); i++) {
            streams.get(i).setOutputIndex(i);
        }
    }

}
